/**
 * Copyright Copyright 2010 Simon Andrews
 *
 *    This file is part of Difference Tracker.
 *
 *    Difference Tracker is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    Difference Tracker is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with Difference Tracker; if not, write to the Free Software
 *    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */
package uk.ac.bbsrc.babraham.DifferenceTracker;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;

import javax.swing.table.TableModel;

import uk.ac.bbsrc.babraham.DifferenceTracker.Config.DifferenceTrackerPreferences;
import uk.ac.bbsrc.babraham.DifferenceTracker.Config.EditTrackerPrefs;
import uk.ac.bbsrc.babraham.DifferenceTracker.Results.ResultsTable;
import uk.ac.bbsrc.babraham.DifferenceTracker.Results.TrackedResultsTableModel;

import ij.IJ;
import ij.ImagePlus;
import ij.Macro;
import ij.plugin.filter.PlugInFilter;
import ij.process.ImageProcessor;
import ij.process.StackConverter;


public class Mass_Particle_Tracker implements PlugInFilter {

	/**
	 * This plugin tries to combine aggregate statistics for a large
	 * number of tracked particles within an image stack.  It is not
	 * concerned with accurately tracking a single particle for a long
	 * time, but more with getting average data for the number of 
	 * particles present, their direction of travel and speed of
	 * movement. 
	 */
	
	public static final String VERSION = "0.1";
	private ImagePlus imp;
	private TrackedParticleCollection trackedCollection = null;
	private boolean unattended = false;
	
//	private static final DecimalFormat df = new DecimalFormat("#.###");
	
	public void run(ImageProcessor processor) {
		
		if (!unattended) {
			EditTrackerPrefs d = new EditTrackerPrefs();
			if (d.wasCancelled()) return;
		}
		
		DifferenceTrackerPreferences prefs = DifferenceTrackerPreferences.getInstance();
		
		int minTrackLength = prefs.minTrackLength();
		
		
		for (int slice = 1;slice<=imp.getStackSize();slice++) {
			IJ.showStatus("Processing slice "+slice);
			ImageProcessor proc = imp.getImageStack().getProcessor(slice);
			ParticleCollection collection = new ParticleCollection(proc);
						
			if (trackedCollection == null) {
				trackedCollection = new TrackedParticleCollection(collection,slice);
			}
			else {
				trackedCollection.extendTrackedParticles(collection, slice);
			}
		}
		
		TrackedParticle [] particleTracks = trackedCollection.getAllTrackedParticles(minTrackLength);
		
		StackConverter converter = new StackConverter(imp);
		converter.convertToRGB();
		
		// Now we go through adding in the tracked regions
		for (int pt=0;pt<particleTracks.length;pt++) {
			particleTracks[pt].addTrackToImagePlus(imp);
		}
		
		ParticleSetStats allParticles = trackedCollection.SummariseMovement(minTrackLength, 0, imp.getStackSize());
		ParticleSetStats leftParticles = trackedCollection.SummariseMovement(minTrackLength, TrackedParticle.LEFT, imp.getStackSize());
		ParticleSetStats rightParticles = trackedCollection.SummariseMovement(minTrackLength, TrackedParticle.RIGHT, imp.getStackSize());
		
		File exportFile = new File(imp.getOriginalFileInfo().directory+"mass_particle_tracker_output.txt");
		
		boolean headersWritten = exportFile.exists();
		
		TableModel resultsModel = new TrackedResultsTableModel(imp.getOriginalFileInfo().fileName,imp,allParticles,leftParticles,rightParticles);
		
		PrintWriter pr = null;
		try {
			pr = new PrintWriter(new FileWriter(exportFile,true));
		} 
		catch (IOException e) {
			e.printStackTrace();
		}

		// We output to the file using the table model.  The model is 2 columns where the
		// first column are the headers and the second column are the values
		
		if (! headersWritten) {
			for (int i=0;i<resultsModel.getRowCount();i++) {
				pr.print(resultsModel.getValueAt(i, 0)); // Column 0 means the field names
				if (i == resultsModel.getRowCount()-1) {
					pr.print("\n");
				}
				else {
					pr.print("\t");
				}
			}			
		}

		// Now we can print out the actual data
		for (int i=0;i<resultsModel.getRowCount();i++) {
			pr.print(resultsModel.getValueAt(i, 1)); // Column 1 means the data
			if (i == resultsModel.getRowCount()-1) {
				pr.print("\n");
			}
			else {
				pr.print("\t");
			}
		}			
		
		pr.close();
		
		imp.updateAndDraw();
		
		if (!unattended) {
			// Don't draw a results table if we're running inside a macro
			new ResultsTable(resultsModel);
		}
		
		if (DifferenceTrackerPreferences.getInstance().produceFullOutput()) {
			try {
				PrintWriter fpr = new PrintWriter(new FileWriter(new File(imp.getOriginalFileInfo().directory+"/"+imp.getOriginalFileInfo().fileName+"_tracked.txt")));

				for (int p=0;p<particleTracks.length;p++) {
					fpr.println("#Particle "+p);
					for (int slice=particleTracks[p].startSlice();slice<=particleTracks[p].endSlice();slice++) {
						XYPosition xy = particleTracks[p].positionInSlice(slice);
						fpr.println(slice+"\t"+xy.x+"\t"+xy.y);
					}
				}
				
				fpr.close();
			}
			
			catch (IOException e) {
				e.printStackTrace();
			}
		}

	}
	

	public int setup(String args, ImagePlus imp) {

		// If they've not passed any arguments then we put up some options.
		// This allows us to run silently inside a macro without pausing
		// to display results.
		
		if (Macro.getOptions() == null) {
			unattended = false;
		}
		else {
			unattended = true;
		}

		this.imp = imp;
		return DOES_8G;
	}

	public static void main (String [] args) {
		new Mass_Particle_Tracker();
	}

}
